/*
 * Copyright (c) 2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Mon Feb 27 2017
 */
#include "acheader.h"

#include <avr/interrupt.h>
#include <avr/io.h>

#include "arduio.h"
#include <util/atomic.h>
#include <util/delay.h>


static enum ario_dir ario_dpins[ARIO_DPINS];
static uint16_t ario_apins[ARIO_APINS];

static uint16_t ario_wspins[ARIO_DPINS][16];

static const uint8_t pwmsibs[14] = {
  [3] = 11, [11] = 3,
  [5] = 6,  [6] = 5,
  [9] = 10, [10] = 9
};

static uint16_t	ocr_scale(uint16_t v, uint16_t s);
static uint16_t	ocr_unscale(uint16_t v, uint16_t s);
static void	timer_prescale(uint8_t n, uint16_t s);


/* --- ario_pin_init ------------------------------------------------------- */

void
ario_pin_init(void)
{
  uint8_t i;

  /* timers at max frequency, phase correct PWM */
  TCCR0A = (0 << WGM01) | (1 << WGM00);
  TCCR0B = (0 << WGM02);
  timer_prescale(0, 1);
  OCR0A = OCR0B = 0;

  TCCR1A = (1 << WGM11) | (0 << WGM10);
  TCCR1B = (1 << WGM13) | (0 << WGM12);
  timer_prescale(1, 1);
  ICR1 = 0xff;
  OCR1A = OCR1B = 0;

  TCCR2A = (0 << WGM21) | (1 << WGM20);
  TCCR2B = (0 << WGM22);
  timer_prescale(2, 1);
  OCR2A = OCR2B = 0;

  /* ADC, AVCC ref, single conversion, 125kHz (~100µs/chan) */
  ADMUX = (0 << REFS1) | (1 << REFS0);
  DIDR0 =
    (1 << ADC5D) | (1 << ADC4D) | (1 << ADC3D) |
    (1 << ADC2D) | (1 << ADC1D) | (1 << ADC0D);
  ADCSRB = (0 << ADTS2) | (0 << ADTS1) | (0 << ADTS0);
  ADCSRA =
    (1 << ADEN) | (1 << ADSC) | (0 << ADATE) | (1 << ADIF) | (1 << ADIE) |
    (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);


  /* input pins by default */
  for(i = 0; i < ARIO_DPINS; i++)
    ario_dpin_dir(i, ARIO_IN);
}


/* --- ario_dpin_dir ------------------------------------------------------- */

void
ario_dpin_dir(uint8_t n, enum ario_dir dir)
{
  if (n >= ARIO_DPINS || n < 2) return;

  /* configure PWM first */
  switch(dir) {
    case ARIO_PWM_31kHz:
      switch(n) {
        case 5:  case 6: timer_prescale(0, 1); break;
        case 9: case 10: timer_prescale(1, 1); ICR1 = 0xff; break;
        case 3: case 11: timer_prescale(2, 1); break;
      }
      break;

    case ARIO_PWM_122Hz:
      switch(n) {
        case 5:  case 6: timer_prescale(0, 256); break;
        case 9: case 10: timer_prescale(1, 1); ICR1 = 0xffff; break;
        case 3: case 11: timer_prescale(2, 256); break;
      }
      break;

    case ARIO_PWM_31Hz:
      switch(n) {
        case 5:  case 6: timer_prescale(0, 1024); break;
        case 9: case 10: timer_prescale(1, 8); ICR1 = 0x7fff; break;
        case 3: case 11: timer_prescale(2, 1024); break;
      }
      break;

    default: break;
  }

  switch(dir) {
    case ARIO_IN:
    case ARIO_INPU:
    case ARIO_OUT:
    case ARIO_WS28XX:
      switch(n) {
        case  3: TCCR2A &= ~((1 << COM2B1) | (0 << COM2B0)); break;
        case  5: TCCR0A &= ~((1 << COM0B1) | (0 << COM0B0)); break;
        case  6: TCCR0A &= ~((1 << COM0A1) | (0 << COM0A0)); break;
        case  9: TCCR1A &= ~((1 << COM1A1) | (0 << COM1A0)); break;
        case 10: TCCR1A &= ~((1 << COM1B1) | (0 << COM1B0)); break;
        case 11: TCCR2A &= ~((1 << COM2A1) | (0 << COM2A0)); break;
      }
      break;

    case ARIO_PWM_31kHz:
    case ARIO_PWM_122Hz:
    case ARIO_PWM_31Hz:
      switch(n) {
        case  3: TCCR2A |= (1 << COM2B1) | (0 << COM2B0); break;
        case  5: TCCR0A |= (1 << COM0B1) | (0 << COM0B0); break;
        case  6: TCCR0A |= (1 << COM0A1) | (0 << COM0A0); break;
        case  9: TCCR1A |= (1 << COM1A1) | (0 << COM1A0); break;
        case 10: TCCR1A |= (1 << COM1B1) | (0 << COM1B0); break;
        case 11: TCCR2A |= (1 << COM2A1) | (0 << COM2A0); break;

        default: return;
      }
      break;
  }

  /* configure direction */
  switch(dir) {
    case ARIO_IN:
      if (n < 8) {	/* portd pins */
        PORTD &= ~(1 << n);
        DDRD &= ~(1 << n);
      } else {		/* portb pins */
        PORTB &= ~(1 << (n-8));
        DDRB &= ~(1 << (n-8));
      }
      break;

    case ARIO_INPU:
      if (n < 8) {	/* portd pins */
        DDRD &= ~(1 << n);
        PORTD |= (1 << n);
      } else {		/* portb pins */
        DDRB &= ~(1 << (n-8));
        PORTB |= (1 << (n-8));
      }
      break;

    case ARIO_OUT:
    case ARIO_PWM_31kHz:
    case ARIO_PWM_122Hz:
    case ARIO_PWM_31Hz:
    case ARIO_WS28XX:
      if (n < 8)	/* portd pins */
        DDRD |= (1 << n);
      else		/* portb pins */
        DDRB |= (1 << (n-8));
      break;
  }

  /* update the frequency of coupled PWM pins */
  switch(dir) {
    case ARIO_PWM_31kHz:
      if (pwmsibs[n]) {
        if (ario_dpins[pwmsibs[n]] > ARIO_OUT &&
            ario_dpins[pwmsibs[n]] < ARIO_WS28XX)
          ario_dpins[pwmsibs[n]] = dir;
      }
      break;

    case ARIO_PWM_122Hz:
      if (pwmsibs[n]) {
        if (ario_dpins[pwmsibs[n]] > ARIO_OUT &&
            ario_dpins[pwmsibs[n]] < ARIO_WS28XX)
          ario_dpins[pwmsibs[n]] = dir;
      }
      break;

    case ARIO_PWM_31Hz:
      if (pwmsibs[n]) {
        if (ario_dpins[pwmsibs[n]] > ARIO_OUT &&
            ario_dpins[pwmsibs[n]] < ARIO_WS28XX)
          ario_dpins[pwmsibs[n]] = dir;
      }
      break;

    default: break;
  }

  ario_dpins[n] = dir;
}


/* --- ario_dpin_out ------------------------------------------------------- */

void
ario_dpin_out(uint8_t n, uint16_t v)
{
  if (n >= ARIO_DPINS || n < 2) return;

  switch(ario_dpins[n]) {
    case ARIO_IN:
    case ARIO_INPU:
      break;

    case ARIO_OUT:
      if (n < 8)	/* portd pins */
        if (v) PORTD |= (1 << n); else PORTD &= ~(1 << n);
      else		/* portb pins */
        if (v) PORTB |= (1 << (n-8)); else PORTB &= ~(1 << (n-8));
      break;

    case ARIO_PWM_31kHz:
    case ARIO_PWM_122Hz:
    case ARIO_PWM_31Hz:
      switch(n) {
        case  3: OCR2B = v >> 8; break;
        case  5: OCR0B = v >> 8; break;
        case  6: OCR0A = v >> 8; break;
        case  9: OCR1A = ocr_scale(v, ICR1); break;
        case 10: OCR1B = ocr_scale(v, ICR1); break;
        case 11: OCR2A = v >> 8; break;

        default: return;
      }
      break;

    case ARIO_WS28XX: {
      uint8_t i;
      uint8_t led = v >> 12;

      ario_wspins[n][led] = v & 0xfff;
      ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
        for(i = 0; i <= led; i++) {
          if (n < 8)
            ws28xx_set(&PORTD, 1 << n, ario_wspins[n][i]);
          else
            ws28xx_set(&PORTB, 1 << (n - 8), ario_wspins[n][i]);
        }
      }
      _delay_us(50);
      break;
    }
  }
}


/* --- ario_dpin_in -------------------------------------------------------- */

uint16_t
ario_dpin_in(uint8_t n)
{
  if (n >= ARIO_DPINS || n < 2) return 0;

  switch(ario_dpins[n]) {
    case ARIO_PWM_31kHz:
    case ARIO_PWM_122Hz:
    case ARIO_PWM_31Hz:
      switch(n) {
        case  3: return OCR2B << 8;
        case  5: return OCR0B << 8;
        case  6: return OCR0A << 8;
        case  9: return ocr_unscale(OCR1A, ICR1);
        case 10: return ocr_unscale(OCR1B, ICR1);
        case 11: return OCR2A << 8;
      }
      /*FALLTHROUGH*/
    case ARIO_IN:
    case ARIO_INPU:
    case ARIO_OUT:
      if (n < 8)	/* portd pins */
        return (PIND & (1 << n)) ? 0xffff : 0;
      else		/* portb pins */
        return (PINB & (1 << (n-8))) ? 0xffff : 0;
      break;

    case ARIO_WS28XX:
      return ario_wspins[n][0];
  }

  return 0;
}


/* --- ario_apin_in -------------------------------------------------------- */

uint16_t
ario_apin_in(uint8_t n)
{
  if (n >= ARIO_APINS) return 0;
  return ario_apins[n];
}


/* --- ario_ddir_state ----------------------------------------------------- */

void
ario_ddir_state(void)
{
  uint8_t dir[1 + ARIO_DPINS];
  uint8_t i;

  dir[0] = 'D';
  for(i = 0; i < sizeof(ario_dpins)/sizeof(ario_dpins[0]); i++)
    dir[i+1] = ario_dpins[i];

  ario_log_buffer(dir, sizeof(dir)/sizeof(dir[0]));
}


/* --- ario_dpin_state ----------------------------------------------------- */

void
ario_dpin_state(void)
{
  uint8_t in[1 + 2 * ARIO_DPINS];
  uint16_t v;
  uint8_t i;

  in[0] = 'I';
  for(i = 0; i < ARIO_DPINS; i++) {
    v = ario_dpin_in(i);
    in[2*i+1] = v >> 8;
    in[2*i+2] = v & 0xff;
  }

  ario_log_buffer(in, sizeof(in));
}


/* --- ario_apin_state ----------------------------------------------------- */

void
ario_apin_state(void)
{
  uint8_t in[1 + 2 * ARIO_APINS];
  uint16_t v;
  uint8_t i;

  in[0] = 'A';
  for(i = 0; i < ARIO_APINS; i++) {
    v = ario_apins[i];
    in[2*i+1] = v >> 8;
    in[2*i+2] = v & 0xff;
  }

  ario_log_buffer(in, sizeof(in));
}


/* --- ocr_scale ----------------------------------------------------------- */

static uint16_t
ocr_scale(uint16_t v, uint16_t s)
{
  switch(s) {
    case 0xffff: return v;
    case 0x7fff: return v >> 1;
    case   0xff: return v >> 8;
  }
  return 0;
}

static uint16_t
ocr_unscale(uint16_t v, uint16_t s)
{
  switch(s) {
    case 0xffff: return v;
    case 0x7fff: return v << 1;
    case   0xff: return v << 8;
  }
  return 0;
}

/* --- timer_prescale ------------------------------------------------------ */

static void
timer_prescale(uint8_t n, uint16_t s)
{
  uint8_t x = 0;

  switch(n) {
    case 0:
      switch(s) {
        case    0: x = (0 << CS02) | (0 << CS01) | (0 << CS00); break;
        case    1: x = (0 << CS02) | (0 << CS01) | (1 << CS00); break;
        case    8: x = (0 << CS02) | (1 << CS01) | (0 << CS00); break;
        case   64: x = (0 << CS02) | (1 << CS01) | (1 << CS00); break;
        case  256: x = (1 << CS02) | (0 << CS01) | (0 << CS00); break;
        case 1024: x = (1 << CS02) | (0 << CS01) | (1 << CS00); break;
      }
      TCCR0B = (TCCR0B & ~((1 << CS02) | (1 << CS01) | (1 << CS00))) | x;
      break;

    case 1:
      switch(s) {
        case    0: x = (0 << CS12) | (0 << CS11) | (0 << CS10); break;
        case    1: x = (0 << CS12) | (0 << CS11) | (1 << CS10); break;
        case    8: x = (0 << CS12) | (1 << CS11) | (0 << CS10); break;
        case   64: x = (0 << CS12) | (1 << CS11) | (1 << CS10); break;
        case  256: x = (1 << CS12) | (0 << CS11) | (0 << CS10); break;
        case 1024: x = (1 << CS12) | (0 << CS11) | (1 << CS10); break;
      }
      TCCR1B = (TCCR1B & ~((1 << CS12) | (1 << CS11) | (1 << CS10))) | x;
      break;

    case 2:
      switch(s) {
        case    0: x = (0 << CS22) | (0 << CS21) | (0 << CS20); break;
        case    1: x = (0 << CS22) | (0 << CS21) | (1 << CS20); break;
        case    8: x = (0 << CS22) | (1 << CS21) | (0 << CS20); break;
        case   32: x = (0 << CS22) | (1 << CS21) | (1 << CS20); break;
        case   64: x = (1 << CS22) | (0 << CS21) | (0 << CS20); break;
        case  128: x = (1 << CS22) | (0 << CS21) | (1 << CS20); break;
        case  256: x = (1 << CS22) | (1 << CS21) | (0 << CS20); break;
        case 1024: x = (1 << CS22) | (1 << CS21) | (1 << CS20); break;
      }
      TCCR2B = (TCCR2B & ~((1 << CS22) | (1 << CS21) | (1 << CS20))) | x;
      break;
  }
}


/* --- ADC_vect ------------------------------------------------------------ */

ISR(ADC_vect)
{
  uint8_t adc_channel = ADMUX & 0x7;

  /* read ADC */
  ario_apins[adc_channel++] = ADC << 6;
  if (adc_channel >= ARIO_APINS) adc_channel = 0;
  ADMUX = (ADMUX & 0xe0) | (adc_channel & 0x7);

  ADCSRA |= (1 << ADSC);
}
